import RoboidControl.Messages.LowLevelMessages as LowLevelMessages
from RoboidControl.Messages.IMessage import IMessage
from LinearAlgebra.Spherical import Spherical
from LinearAlgebra.Quaternion import Quaternion
from RoboidControl.Thing import Thing

from typing import Union, Optional

class PoseMsg(IMessage):
    id = 0x10
    length = 4

    Position = 0x01
    Orientation = 0x02    
    LinearVelocity = 0x04
    AngularVelocity = 0x08

    def __init__(self, arg1: Union[bytes, int], thing: Optional[Thing], force: bool = False):
        if thing is None:
            buffer = bytes(arg1)
            self.network_id = buffer[1]
            self.thing_id = buffer[2]
            self.pose_type = buffer[3]

            self.position: Optional[Spherical] = None
            self.orientation: Optional[Quaternion] = None
            self.linear_velocity: Optional[Spherical] = None
            self.angular_velocity: Optional[Spherical] = None

            ix:int = 4
            ix_ref = [ix]
            if self.pose_type & PoseMsg.Position != 0:
                self.position = LowLevelMessages.ReceiveSpherical(buffer, ix_ref)
            if self.pose_type & PoseMsg.Orientation != 0:
                self.orientation = LowLevelMessages.ReceiveQuaternion(buffer, ix_ref)
            if self.pose_type & PoseMsg.LinearVelocity != 0:
                self.linear_velocity = LowLevelMessages.ReceiveSpherical(buffer, ix_ref)
            if self.pose_type & PoseMsg.AngularVelocity != 0:
                self.angular_velocity = LowLevelMessages.ReceiveSpherical(buffer, ix_ref)
        else:
            self.network_id = arg1
            self.thing = thing

            self.pose_type = 0
            if thing.position_updated or force:
                self.pose_type |= PoseMsg.Position
            if thing.orientation_updated or force:
                self.pose_type |= PoseMsg.Orientation
            if thing.linear_velocity_updated or force:
                self.pose_type |= PoseMsg.LinearVelocity
            if thing.angular_velocity_updated or force:
                self.pose_type |= PoseMsg.AngularVelocity

    def Serialize(self, buffer_ref):
        if self.thing is None or self.pose_type == 0:
            return 0
        
        print(f'Send PoseMsg [{self.network_id}/{self.thing.id}] {self.pose_type}')

        buffer: bytearray = buffer_ref[0]
        buffer[0:PoseMsg.length] = [
            PoseMsg.id,
            self.network_id,
            self.thing.id,
            self.pose_type
        ]
        ix = [4]
        if self.pose_type & PoseMsg.Position:
            LowLevelMessages.SendSpherical(buffer, ix, self.thing.position)
        if self.pose_type & PoseMsg.Orientation:
            LowLevelMessages.SendQuat32(buffer, ix, self.thing.orientation)
        if self.pose_type & PoseMsg.LinearVelocity:
            LowLevelMessages.SendSpherical(buffer, ix, self.thing.linear_velocity)
        if self.pose_type & PoseMsg.AngularVelocity:
            LowLevelMessages.SendSpherical(buffer, ix, self.thing.angular_velocity)
        return PoseMsg.length + ix[0]
